// =====================================
// 1. Imports and Configurations
// =====================================
import 'dotenv/config';
import express from 'express';
import expressWs from 'express-ws';
import twilio from 'twilio';
import { ElevenLabsClient } from 'elevenlabs';
import WebSocket from 'ws';
import { spawn } from 'child_process';
import { Readable } from 'stream';
import path from 'path';
import fs from 'fs/promises';
import OpenAI from 'openai';
import { createClient, LiveTranscriptionEvents, DeepgramClient, DeepgramError } from '@deepgram/sdk';
import Groq from 'groq-sdk';
import { Buffer } from 'buffer';
import { EventEmitter } from 'events';
import webhookRouter from './webhook';
import VoiceResponse = require('twilio/lib/twiml/VoiceResponse');
import crypto from 'crypto';

// Global variables
let calledNumber = '';
let conversationSummary = '';
let summaryAudioBuffer: Buffer | null = null;
let ID = '';
let From = '';
let To = '';

// Configuration
type OpenAIVoice = "alloy" | "echo" | "fable" | "onyx" | "nova" | "shimmer";

interface AppConfig {
  enable_interruptions: boolean;
  wordsToInterrupt: string[];
  wordToInterruptEnable: boolean;
  numberofwordstointeruptenable: boolean;
  words_to_interrupt: number;
  silenceTimeoutSeconds: number;
  response_delay: number;
  llm_request_delay: number;
  clear_message_delay: number;
  transcriber_provider: string;
  transcriber_model: string;
  transcriber_language: string;
  model_provider: string;
  model: string;
  temperature: number;
  top_p: number;
  max_tokens: number;
  voice_provider: string;
  voice_id: string | OpenAIVoice;
  stability: number;
  clarity_boost: number;
  speed: number;
  maximum_call_duration: number;
  maximum_call_duration_enable: boolean;
  hipaa_compliance: boolean;
  recording_enabled: boolean;
  server_url: string;
  voicemail_message: string;
  end_call_message: string;
  call_forwarding_enable: boolean;
  forwarding_phone_number: string;
  call_forwarding_confirmation_message: string;
  enable_sms_booking_text: boolean;
  sms_booking_text: string;
  system_prompt: string;
  call_forwarding_mode: 'Direct' | 'Assisted';
  assistant_name: string;
  endpointing_ms: number;
  utterance_end_ms: number;
  utterance_timeout_ms: number;
  provider_settings?: {
    openai_api_key: string;
    groq_api_key: string;
    elevenlabs_api_key: string;
    rime_ai_api_key: string;
    deepgram_api_key: string;
    twilio_account_sid: string;
    twilio_auth_token: string;
  };
}

const appConfig: AppConfig = {
  enable_interruptions: false, // Default value, will be overridden by assistant settings if available
  wordsToInterrupt: ['stop', 'wait', 'hold on'],
  wordToInterruptEnable: false,
  numberofwordstointeruptenable: true,
  words_to_interrupt: 3, // Default value, will be overridden by assistant settings if available
  silenceTimeoutSeconds: 30, // Default value, will be overridden by assistant settings if available
  response_delay: 0.1, // Default value, will be overridden by assistant settings if available
  llm_request_delay: 0.1, // Default value, will be overridden by assistant settings if available
  clear_message_delay: 500, // Default value, will be overridden by assistant settings if available
  transcriber_provider: 'deepgram', // Default value, will be overridden by assistant settings if available
  transcriber_model: 'nova-2', // Default value, will be overridden by assistant settings if available
  transcriber_language: 'nova-2', // Default value, will be overridden by assistant settings if available
  model_provider: 'openai', // Default value, will be overridden by assistant settings if available
  model: 'gpt-3.5-turbo', // Default value, will be overridden by assistant settings if available
  temperature: 0.7, // Default value, will be overridden by assistant settings if available
  top_p: 1, // Default value, will be overridden by assistant settings if available
  max_tokens: 150, // Default value, will be overridden by assistant settings if available
  voice_provider: 'elevenlabs', // Default value, will be overridden by assistant settings if available
  voice_id: 'BJStPCleYBG69brKsxqv', // Default value, will be overridden by assistant settings if available
  stability: 0.5, // Default value for ElevenLabs, will be overridden by assistant settings if available
  clarity_boost: 0.75, // Default value for ElevenLabs, will be overridden by assistant settings if available
  speed: 1.0, // Default value, will be overridden by assistant settings if available
  maximum_call_duration: 600, // Default value, 600 seconds (10 minutes)
  maximum_call_duration_enable: false, // Default value, will be overridden by assistant settings if available
  hipaa_compliance: true, // to implement
  recording_enabled: true, // to implement
  server_url: "http://", // Default value, should be updated with actual server URL. to implement
  voicemail_message: "You have reached our voicemail. Please leave a message after the tone.", // to implement
  end_call_message: "Goodbye", // to implement
  call_forwarding_enable: false, // Default value for call forwarding
  forwarding_phone_number: "", // Default value for forwarding phone number
  call_forwarding_confirmation_message: "", // Default value for forwarding phone number
  enable_sms_booking_text: true, // Default value for SMS booking feature
  sms_booking_text: "", // Default value for SMS booking text
  system_prompt: "",
  call_forwarding_mode: 'Assisted', // Default to Assisted mode | Direct
  assistant_name:'Assistant', // Default Assistant name
  endpointing_ms: 300, // Default value for Deepgram endpointing in milliseconds
  utterance_end_ms: 1000, // 1200 Default value for Deepgram utterance_end_ms in milliseconds
  utterance_timeout_ms: 1200, // 2000 Default value for utterance timeout in milliseconds
};

async function getAssistantSettings(phoneNumber: string) {
  log(`Fetching assistant settings for phone number: ${phoneNumber}`);
  const cachePath = path.join(__dirname, '..', 'cache_settings', 'assistantSettings.json');
  try {
    const fileContent = await fs.readFile(cachePath, 'utf8');
    const assistants = JSON.parse(fileContent);
    const settings = assistants.find((a: any) => a.phone_number === phoneNumber);
    if (settings) {
      log(`Found assistant settings for ${phoneNumber}:`, settings);
      
      // Update appConfig with the fetched settings
      if (settings.transcriber_provider) {
        appConfig.transcriber_provider = settings.transcriber_provider;
      }
      if (settings.transcriber_model) {
        appConfig.transcriber_model = settings.transcriber_model;
      }
      if (settings.transcriber_language) {
        appConfig.transcriber_language = settings.transcriber_language;
      }
      if (settings.silence_timeout) {
        appConfig.silenceTimeoutSeconds = parseInt(settings.silence_timeout);
      }
      if (settings.model_provider) {
        appConfig.model_provider = settings.model_provider;
      }
      if (settings.model) {
        appConfig.model = settings.model;
      }
      if (settings.temperature) {
        appConfig.temperature = parseFloat(settings.temperature);
      }
      if (settings.top_p) {
        appConfig.top_p = parseFloat(settings.top_p);
      }
      if (settings.max_tokens) {
        appConfig.max_tokens = parseInt(settings.max_tokens);
      }
      if (settings.voice_provider) {
        appConfig.voice_provider = settings.voice_provider;
      }
      if (settings.voice_id) {
        appConfig.voice_id = settings.voice_id;
      }
      if (settings.stability && settings.voice_provider === 'elevenlabs') {
        appConfig.stability = parseFloat(settings.stability);
      }
      if (settings.clarity_boost && settings.voice_provider === 'elevenlabs') {
        appConfig.clarity_boost = parseFloat(settings.clarity_boost);
      }
      if (settings.speed) {
        appConfig.speed = parseFloat(settings.speed);
      }
      if (settings.response_delay) {
         appConfig.response_delay = parseFloat(settings.response_delay);
      }
      if (settings.llm_request_delay) {
         appConfig.llm_request_delay = parseFloat(settings.llm_request_delay);
      }
      if (settings.clear_message_delay) {
         appConfig.clear_message_delay = parseFloat(settings.clear_message_delay);
      }
      if (settings.enable_interruptions) {
       appConfig.enable_interruptions = Boolean(settings.enable_interruptions);
      }
      if (settings.words_to_interrupt) {
         appConfig.words_to_interrupt = parseInt(settings.words_to_interrupt);
      }
      if (settings.call_forwarding_enable !== undefined) {
        appConfig.call_forwarding_enable = Boolean(settings.call_forwarding_enable);
      }
      if (settings.call_forwarding_mode) {
       appConfig.call_forwarding_mode = settings.call_forwarding_mode;
      }
      if (settings.forwarding_phone_number) {
        appConfig.forwarding_phone_number = settings.forwarding_phone_number;
      }
      if (settings.call_forwarding_confirmation_message) {
        appConfig.call_forwarding_confirmation_message = settings.call_forwarding_confirmation_message;
      }
      if (settings.enable_sms_booking_text !== undefined) {
        appConfig.enable_sms_booking_text = Boolean(settings.enable_sms_booking_text);
      }
      if (settings.sms_booking_text) {
        appConfig.sms_booking_text = settings.sms_booking_text;
      }
      if (settings.system_prompt) {
        appConfig.system_prompt = settings.system_prompt;
      }
      if (settings.assistant_name) {
        appConfig.assistant_name = settings.assistant_name;
      }
      if (settings.endpointing_ms) {
        appConfig.endpointing_ms = parseInt(settings.endpointing_ms);
      }
      if (settings.utterance_end_ms) {
        appConfig.utterance_end_ms = parseInt(settings.utterance_end_ms);
      }
      if (settings.utterance_timeout_ms) {
        appConfig.utterance_timeout_ms = parseInt(settings.utterance_timeout_ms);
      }
      
      return settings;
    } else {
      log(`No assistant settings found for phone number: ${phoneNumber}`);
      return null;
    }
  } catch (error) {
    log('Error reading assistantSettings.json:', error);
    return null;
  }
}

// The fetchAssistantSettings function has been moved to appConfig.ts
// Import and use it from there when needed
import { fetchAssistantSettings } from './config/appConfig';

// Function to set the LLM API type based on the model provider
function setLLMApiType(modelProvider: string) {
  switch (modelProvider.toLowerCase()) {
    case 'openai':
      CURRENT_LLM_API = LLMApiType.OpenAIChat;
      break;
    case 'groq':
      CURRENT_LLM_API = LLMApiType.Groq;
      break;
    default:
      log('Unknown model provider, defaulting to OpenAIChat', { provider: modelProvider });
      CURRENT_LLM_API = LLMApiType.OpenAIChat;
  }
  log('Set LLM API Type', { provider: modelProvider, apiType: CURRENT_LLM_API });
}


// Initialize express app
const { app } = expressWs(express());
const PORT: number = parseInt(process.env.PORT || '3000', 10);

// Declare API client variables
let elevenlabs: ElevenLabsClient;
let openai: OpenAI;
let deepgram: DeepgramClient;
let groq: Groq;
let twilioClient: ReturnType<typeof twilio>;

// Function to initialize API clients with current settings
function initializeAPIClients(settings: AppConfig) {
    if (!settings.provider_settings) {
        throw new Error('Provider settings not found in assistant settings');
    }
    
    elevenlabs = new ElevenLabsClient({ apiKey: settings.provider_settings.elevenlabs_api_key });
    openai = new OpenAI({ apiKey: settings.provider_settings.openai_api_key });
    deepgram = createClient(settings.provider_settings.deepgram_api_key);
    groq = new Groq({ apiKey: settings.provider_settings.groq_api_key });
    twilioClient = twilio(
        settings.provider_settings.twilio_account_sid,
        settings.provider_settings.twilio_auth_token
    );
    
    log('API clients initialized with current provider settings');
}

let isFirstMessage = true;
let silenceTimer: NodeJS.Timeout | null = null;
let ws: WebSocket;
let isAssistantSpeaking = false;
let isWaitingForAudioComplete = false;

const RIME_AI_CONFIG = {
  speaker: process.env.RIME_AI_SPEAKER || 'default_speaker_name',
  modelId: process.env.RIME_AI_MODEL || 'mist',
  reduceLatency: process.env.RIME_AI_REDUCE_LATENCY === 'true'
};

// =====================================
// 1.5 Global State
// =====================================
let currentAssistantSettings: any = null;

// =====================================
// 2. Type Definitions and Interfaces
// =====================================
enum LLMApiType {
  OpenAIAssistant,
  OpenAIChat,
  Groq
}

enum TTSProvider {
  ElevenLabs,
  OpenAI,
  Deepgram,
  RimeAI
}

enum CommunicationMethod {
  NonStreaming,
  Streaming,
  WebSocket
}

type AssistantStreamEvent = {
  event: string;
  data: any;
};

interface TwilioMessage {
  event: string;
  streamSid?: string;
  sequenceNumber?: string;
  start?: {
    streamSid: string;
    callSid: string;
    accountSid: string;
    tracks: string[];
    customParameters: Record<string, string>;
    mediaFormat: {
      encoding: string;
      sampleRate: number;
      channels: number;
    };
  };
  media?: {
    track: string;
    chunk: string;
    timestamp: string;
    payload: string;
  };
  stop?: {
    accountSid: string;
    callSid: string;
  };
  dtmf?: {
    track: string;
    digit: string;
  };
  mark?: {
    name: string;
  };
  error?: string; // Add this line to include the error property
}

interface ConversationMessage {
  role: 'user' | 'assistant';
  content: string;
  timestamp?: string;
}

type ChatMessage = {
  role: 'system' | 'user' | 'assistant';
  content: string;
  timestamp?: string;
};

interface CallAnalysis {
  sentiment: string;
  speaking_time: {
    ai: number;
    user: number;
  };
  key_topics: string[];
  tags: string[];
  latencies: {
    [key: string]: number;
  };
}

interface CallLog {
  id: string;
  phone: string;
  assistant_name: string;
  duration: string;
  timestamp: string;
  summary: string;
  status: string;
  tags: string[];
  model: string;
  model_provider: string;
  voice_provider: string;
  transcriber_provider: string;
  transcript: {
    role: 'assistant' | 'user';
    message: string;
    time: string;
  }[];
  analysis: CallAnalysis;
}

type MessageContentText = {
  type: 'text';
  text: {
    value: string;
    annotations: any[];
  };
};

type MessageContentImage = {
  type: 'image_file';
  image_file: {
    file_id: string;
  };
};

type MessageContent = MessageContentText | MessageContentImage;

// =====================================
// 3. Global Variables and Constants
// =====================================
let isSpeaking = false;
const interruptionEmitter = new EventEmitter();
let mediaSequenceNumber = 0;
let CURRENT_LLM_API: LLMApiType = LLMApiType.Groq;
let isProcessing = false;
const CURRENT_TTS_PROVIDER: TTSProvider = TTSProvider.ElevenLabs;
const COMMUNICATION_METHOD: CommunicationMethod = CommunicationMethod.Streaming;
const ffmpegPath = path.join(__dirname, '..', 'third-party', 'ffmpeg', 'ffmpeg');
let callSid: string | undefined;
let callStartTime: number | null = null;

function calculateCallDuration(): string {
  if (!callStartTime) return "0:00";
  const duration = Math.floor((Date.now() - callStartTime) / 1000);
  const minutes = Math.floor(duration / 60);
  const seconds = duration % 60;
  return `${minutes}:${seconds.toString().padStart(2, '0')}`;
}

const voiceId = 'BJStPCleYBG69brKsxqv';
const modelId = 'eleven_flash_v2_5';
let currentThreadId: string | null = null;
const USE_STREAMING = false;

const GENERAL_INSTRUCTION = `You are a conversational voice assistant for Blinks Media. Use short, natural responses under 20 words as if you're having a live phone conversation. Be friendly, helpful, and concise.`;
 
const BOOKING_INSTRUCTION = ` ## BOOKING INSTRUCTION
- When user shows interest in booking an appointment:
     * First confirm: "Just to confirm, you want to book an appointment, right?"
     * Upon confirmation, respond ONLY with: "Send SMS booking text."
     * After SMS is sent, shift focus to answering questions or providing support
     * Do not push or ask them to book again
     * If appropriate, provide a gentle reminder about using the link
     * Thank them if they confirm they've completed the booking process`;
 
const CALL_FORWARDING_INSTRUCTION = `## HUMAN ESCALATION PROTOCOL
- When customer expresses ANY desire to speak with a human or have their call forwarded:
  * First confirm: "Just to confirm, you'd like me to forward your call to a human representative, is that correct?"
  * Upon confirmation, respond ONLY with: "Forwarding your call now."
  * Follow this procedure whenever human escalation sentiment is detected, regardless of phrasing`; 
 
const COMMUNICATION_STYLE = `## COMMUNICATION STYLE
- Be conversational, friendly, and empathetic
- Avoid sounding robotic or generic
- Make the prospect feel at ease before diving into questions
- Gently guide the conversation back to booking appointments if it strays
- Keep our process confidential (do not explain your internal workflows)
- Use positive language that emphasizes solutions
- Mirror the customer's communication style and energy level when appropriate
- Incorporate brief pauses ("hmm," "let me see") to sound more natural
- Ask clarifying questions rather than making assumptions
- Acknowledge concerns or objections before addressing them
- Use light humor when appropriate, but remain professional
- Personalize responses by referencing previous parts of the conversation
- Express genuine interest in the customer's business needs
- Use simple, jargon-free language unless the customer demonstrates industry knowledge
- End interactions with clear next steps or a summary of what was accomplished`;
 
const RESPONSE_GUIDELINES = `## RESPONSE GUIDELINES
- **Be concise:** 
  * Address one topic at a time
  * Keep responses under 160 characters
  * Avoid repetition and unnecessary details

- **Embrace variety:** 
  * Use diverse language and rephrasing
  * Employ everyday conversational language
  * Make interactions feel like talking to a friend

- **Be proactive:** 
  * Lead the conversation
  * End with a single question or next-step suggestion
  * Never ask multiple questions in one response

- **Get clarity:** 
  * Follow up when answers are partial or unclear
  * Persist politely until you have complete information

- **Adapt and interpret:** 
  * Work with potential transcription errors
  * Never mention "transcription error" in responses

- **Stay on topic:** 
  * Guide conversation back on track when it veers off
  * Continue from current point without restarting

- **Stay in character:** 
  * Maintain your role's scope
  * Guide conversations back creatively without repetition

- **Ensure fluid dialogue:** 
  * Respond in a role-appropriate, direct manner
  * Use colloquial date references like: "next Friday," "tomorrow"

- **Use approved content:** 
  * Never fabricate responses
  * Acknowledge uncertainty and refocus on relevant topics

- **Avoid emojis and unsolicited promotions:** 
  * Do not use emojis under any circumstances
  * Discuss services only when interest is shown
  * Confirm interest before offering additional help

- **Personal touch:** 
  * Ask for first name early in conversation
  * Use name sparingly to maintain natural tone

- **Non-service questions:** 
  * Acknowledge off-topic questions briefly
  * Suggest online resources
  * Refocus on services: "Interesting question! While it's outside our focus, online search might help. How can we assist with our services?"

- **Spam handling:** 
  * Acknowledge politely but discourage engagement
  * Example: "Thanks for reaching out. We're not interested in new opportunities right now. Best wishes!"

- **One question at a time:** 
  * Ask exactly one question per response
  * Never combine multiple topics in one message
  
  ## TTS OPTIMIZATION
- Use proper punctuation to guide speech pacing and intonation
- Spell out numbers, dates, URLs, and email addresses for clear pronunciation
- Avoid unusual characters, symbols or abbreviations that may be misread
- Use phonetic spellings for uncommon words or names in parentheses when needed
- Insert subtle pauses with commas where natural breaks should occur
- Use contractions to maintain conversational flow like: don't, can't, we'll
- Avoid special formatting like ALL CAPS that might affect pronunciation
- For critical information, space out words or use hyphens like: "A-B-C" instead of "ABC"
- Spell check all responses to prevent TTS mispronunciations`;
 
// Get System Prompt
 function getCombinedSystemPrompt(): string {
    let combinedPrompt = `${GENERAL_INSTRUCTION}

 ${BOOKING_INSTRUCTION}

 ${CALL_FORWARDING_INSTRUCTION}

 ${COMMUNICATION_STYLE}

 ${RESPONSE_GUIDELINES}

 ${appConfig.system_prompt}`;

    if (appConfig.enable_interruptions) {
      combinedPrompt += `

 Interruption Handling:
 - If the user interrupts, acknowledge it politely (e.g., "I see you have something to add," or "Oh, sorry for 
 interrupting. You were saying?").
 - After acknowledging, address their new input directly.
 - Keep the acknowledgment brief and natural, then focus on responding to the new information or question.
 `;
    }

    return combinedPrompt;
 }

let chatHistory: ChatMessage[] = [];

function initializeChatHistory() {
   chatHistory = [
     { role: 'system', content: getCombinedSystemPrompt() }
   ];
 }

// =====================================
// 4. Utility Functions
// =====================================
function delay(ms: number): Promise<void> {
  return new Promise(resolve => setTimeout(resolve, ms));
}

function log(message: string, data?: any) {
  const timestamp = new Date().toISOString();
  let logMessage = `${timestamp}: ${message}`;
  if (data) {
    logMessage += ` ${JSON.stringify(data)}`;
  }
  console.log(logMessage);
}

function switchLLMProvider(provider: LLMApiType) {
  CURRENT_LLM_API = provider;
  log('Switched LLM provider', { newProvider: provider });
}

function clearChatHistory() {
   initializeChatHistory();
   log('Chat history cleared and reinitialized');
 }

function sendClearMessage(ws: WebSocket, streamSid: string) {
  const clearMessage = {
    event: "clear",
    streamSid: streamSid
  };
  log('Sending clear message to Twilio', { message: clearMessage });
  ws.send(JSON.stringify(clearMessage));
}

// =====================================
// 5. LatencyTracker Class
// =====================================
class LatencyTracker {
  private latencies: { [key: string]: number[] } = {};
  private startTimes: { [key: string]: number } = {};

  startTracking(service: string) {
    this.startTimes[service] = Date.now();
  }

  endTracking(service: string) {
    const endTime = Date.now();
    const startTime = this.startTimes[service];
    if (startTime) {
      const latency = endTime - startTime;
      if (!this.latencies[service]) {
        this.latencies[service] = [];
      }
      this.latencies[service].push(latency);
      delete this.startTimes[service];
      this.logLatency(service, latency);
    }
  }

  private logLatency(service: string, latency: number) {
    console.log(`${service} Latency: ${latency}ms`);
  }

  getAverages() {
    const averages: { [key: string]: number } = {};
    for (const [service, latencies] of Object.entries(this.latencies)) {
      averages[service] = latencies.reduce((a, b) => a + b, 0) / latencies.length;
    }
    return averages;
  }

  logSummary() {
    const averages = this.getAverages();
    console.log('Call Latency Summary', { averageLatencies: averages });
  }

  getLatency(service: string): number | undefined {
    const latencies = this.latencies[service];
    if (latencies && latencies.length > 0) {
      return latencies[latencies.length - 1];
    }
    return undefined;
  }
}

 async function forwardCall(ws: WebSocket, streamSid: string, callSid: string) {
   if (!callSid) {
     log('Error: CallSid is not available');
     const errorMessage = {
       event: 'call_forward_error',
       streamSid: streamSid,
       error: 'CallSid is not available'
     };
     ws.send(JSON.stringify(errorMessage));
     return;
   }

   log('Forwarding call', { forwardingNumber: appConfig.forwarding_phone_number, callSid, callerId: calledNumber });      

   if (!currentAssistantSettings?.provider_settings) {
     log('Error: Provider settings not available');
     throw new Error('Provider settings not available');
   }

   const accountSid = currentAssistantSettings.provider_settings.twilio_account_sid;
   const authToken = currentAssistantSettings.provider_settings.twilio_auth_token;
   const client = twilio(accountSid, authToken);

   try {
     const twiml = new twilio.twiml.VoiceResponse();
     
     if (appConfig.call_forwarding_mode === 'Assisted') {
       // Generate and use the summary only in Assisted mode
       conversationSummary = await generateConversationSummary();
       twiml.say('Forwarding your call now. Please hold while I brief the representative.');
       twiml.pause({ length: 1 });
     } else {
       // In Direct mode, just forward the call without a summary
       twiml.say('Forwarding your call now. Please hold.');
     }

     twiml.dial({
       callerId: calledNumber,
       answerOnBridge: true
     }).number({
       url: appConfig.call_forwarding_mode === 'Assisted' ? `https://${process.env.SERVER_DOMAIN}:8443/call/whisper` :    
 undefined
     }, appConfig.forwarding_phone_number);

     const twimlString = twiml.toString();

     const call = await client.calls(callSid).update({
       twiml: twimlString
     });

     log('Call forwarded successfully', { callSid: call.sid, callerId: calledNumber });

     const message = {
       event: 'call_forwarded',
       streamSid: streamSid,
       callSid: call.sid
     };
     ws.send(JSON.stringify(message));
   } catch (error) {
     log('Error forwarding call', { error: (error as Error).message });
     const errorMessage = {
       event: 'call_forward_error',
       streamSid: streamSid,
       callSid: callSid,
       error: (error as Error).message
     };
     ws.send(JSON.stringify(errorMessage));
   }
 }

async function generateConversationSummary(): Promise<string> {
  const summaryPrompt = `Please provide a brief summary of the conversation so far, highlighting the main points and any important information discussed. Add at the end of the summary this: ${appConfig.call_forwarding_confirmation_message}.`;
  const summaryResponse = await getLLMResponse(summaryPrompt, null, new LatencyTracker(), CURRENT_LLM_API);
  return typeof summaryResponse === 'string' ? summaryResponse : 'Unable to generate summary';
}

async function generateAudioBuffer(text: string): Promise<Buffer> {
   const latencyTracker = new LatencyTracker();
   latencyTracker.startTracking('TTS');

   let audioBuffer: Buffer;

   switch (appConfig.voice_provider.toLowerCase()) {
 case 'elevenlabs':
   log('Starting ElevenLabs TTS process');
   try {
     const url = `https://api.elevenlabs.io/v1/text-to-speech/${appConfig.voice_id}`;
     
     const apiKey = currentAssistantSettings.provider_settings.elevenlabs_api_key;
     if (!apiKey) {
       throw new Error('ELEVENLABS_API_KEY is not set in the environment variables');
     }

     const headers: HeadersInit = {
       "Accept": "audio/mpeg",
       "Content-Type": "application/json",
       "xi-api-key": currentAssistantSettings.provider_settings.elevenlabs_api_key
     };

     const data = {
       text: text,
       model_id: modelId,
       voice_settings: {
         stability: appConfig.stability,
         similarity_boost: appConfig.clarity_boost
       }
     };

     const response = await fetch(url, {
       method: 'POST',
       headers: headers,
       body: JSON.stringify(data)
     });

     if (!response.ok) {
       throw new Error(`ElevenLabs API responded with status: ${response.status}`);
     }

     if (!response.body) {
       throw new Error('No response body received from ElevenLabs');
     }

     const chunks: Uint8Array[] = [];
     const reader = response.body.getReader();

     while (true) {
       const { done, value } = await reader.read();
       if (done) break;
       chunks.push(value);
     }

     audioBuffer = Buffer.concat(chunks);
     log('Successfully processed ElevenLabs response', { bufferLength: audioBuffer.length });
   } catch (error) {
     log('Error in ElevenLabs TTS', { error: (error as Error).message });
     throw error;
   }
   break;

 case 'openai':
   const mp3Audio = await openai.audio.speech.create({
     model: "tts-1",
     voice: getValidOpenAIVoice(appConfig.voice_id),
     input: text,
   });

   log('OpenAI response type:', typeof mp3Audio);

   if (mp3Audio && typeof mp3Audio.arrayBuffer === 'function') {
     try {
       const arrayBuffer = await mp3Audio.arrayBuffer();
       audioBuffer = Buffer.from(arrayBuffer);
       log('Successfully converted OpenAI response to Buffer', { bufferLength: audioBuffer.length });
     } catch (error) {
       log('Error converting OpenAI response to Buffer:', error);
       throw new Error('Unable to convert OpenAI response to Buffer');
     }
   } else {
     log('Unexpected OpenAI response structure:', mp3Audio);
     throw new Error('Unexpected response structure from OpenAI');
   }
   break;

     case 'deepgram':
       const response = await deepgram.speak.request(
         { text: text },
         {
           model: appConfig.voice_id,
           encoding: "mp3",
         }
       );
       log('Deepgram response Summary');
       const stream = await response.getStream();
       if (!stream) {
         throw new Error('No audio stream received from Deepgram');
       }
       audioBuffer = await streamToBuffer(stream);
       break;

     case 'rime_ai':
   log('Starting RimeAI TTS process');
   try {
     const url = 'https://users.rime.ai/v1/rime-tts';
     
     if (!currentAssistantSettings?.provider_settings?.rime_ai_api_key) {
       throw new Error('Rime AI API key not found in provider settings');
     }

     const headers: HeadersInit = {
       'Accept': 'audio/mp3',
       'Authorization': `Bearer ${currentAssistantSettings.provider_settings.rime_ai_api_key}`,
       'Content-Type': 'application/json'
     };

     const data = {
       speaker: appConfig.voice_id,
       text: text,
       modelId: RIME_AI_CONFIG.modelId,
       speedAlpha: appConfig.speed,
       reduceLatency: RIME_AI_CONFIG.reduceLatency
     };

     const response = await fetch(url, {
       method: 'POST',
       headers: headers,
       body: JSON.stringify(data)
     });

     if (!response.ok) {
       throw new Error(`RimeAI API responded with status: ${response.status}`);
     }

     if (!response.body) {
       throw new Error('No response body received from RimeAI');
     }

     const chunks: Uint8Array[] = [];
     const reader = response.body.getReader();

     while (true) {
       const { done, value } = await reader.read();
       if (done) break;
       chunks.push(value);
     }

     audioBuffer = Buffer.concat(chunks);
     log('Successfully processed RimeAI response', { bufferLength: audioBuffer.length });
   } catch (error) {
     log('Error in RimeAI TTS', { error: (error as Error).message });
     throw error;
   }
   break;

     default:
       throw new Error(`Unsupported voice provider: ${appConfig.voice_provider}`);
   }

   latencyTracker.endTracking('TTS');
   return audioBuffer;
 }

async function streamToBuffer(stream: ReadableStream<Uint8Array> | null): Promise<Buffer> {
   if (!stream) {
     throw new Error('Stream is null');
   }
   const chunks: Uint8Array[] = [];
   const reader = stream.getReader();
   while (true) {
     const { done, value } = await reader.read();
     if (done) break;
     chunks.push(value);
   }
   return Buffer.concat(chunks);
 }

// =====================================
// 6. TranscriptionService Class
// =====================================
class TranscriptionService {
  private dgConnection: ReturnType<DeepgramClient['listen']['live']>;
  private keepAliveInterval: NodeJS.Timeout | null = null;
  private reconnectTimeout: NodeJS.Timeout | null = null;
  private utteranceTimeout: NodeJS.Timeout | null = null;
  private finalResult: string = '';
  private speechFinal: boolean = false;
  private isActive: boolean = true;
  private reconnectAttempts: number = 0;
  private maxReconnectAttempts: number = 5;
  private transcriptionStartTime: number = 0;
  private latencyTracker: LatencyTracker;
  private interruptionDetected: boolean = false;
  private enableInterruptions: boolean = false;
  private ws: WebSocket;
  private streamSid: string;
  private lastWordTimestamp: number = 0;
  private isProcessingTranscription: boolean = false;
  private lastProcessedText: string = '';

  constructor(
    private handleTranscription: (transcription: string, wasInterrupted: boolean) => void,
    latencyTracker: LatencyTracker,
    enable_interruptions: boolean = false,
    ws: WebSocket,
    streamSid: string
  ) {
    this.latencyTracker = latencyTracker;
    this.enableInterruptions = enable_interruptions;
    this.dgConnection = this.initializeConnection();
    this.ws = ws;
    this.streamSid = streamSid;
  }

  private initializeConnection(): ReturnType<DeepgramClient['listen']['live']> {
    const connection = deepgram.listen.live({
      encoding: 'mulaw',
      sample_rate: 8000,
      model: appConfig.transcriber_model,
      language: appConfig.transcriber_language,
      smart_format: true,
      punctuate: true,
      interim_results: true,
      no_delay: true,
      endpointing: appConfig.endpointing_ms,
      utterance_end_ms: appConfig.utterance_end_ms,
      vad_events: true
    });

    this.setupEventListeners(connection);
    this.startKeepAlive();

    return connection;
  }

  private setupEventListeners(connection: ReturnType<DeepgramClient['listen']['live']>) {
    connection.on(LiveTranscriptionEvents.Open, () => {
      log('Deepgram connection opened');
      this.reconnectAttempts = 0;
    });

    connection.on(LiveTranscriptionEvents.Close, () => {
      log('Deepgram connection closed');
      this.stopKeepAlive();
      this.attemptReconnection();
    });

    connection.on(LiveTranscriptionEvents.Transcript, this.handleTranscriptEvent.bind(this));

    connection.on(LiveTranscriptionEvents.Error, (error) => {
      log('Deepgram error:', error);
    });
  }

  private async handleTranscriptEvent(transcriptionEvent: any) {
    // If we're already processing a transcription, don't start another one
    if (this.isProcessingTranscription) {
      log('Already processing a transcription, skipping this event');
      return;
    }

    const alternatives = transcriptionEvent.channel?.alternatives;
    let text = alternatives?.[0]?.transcript || '';

    // Handle UtteranceEnd event
    if (transcriptionEvent.type === 'UtteranceEnd') {
      this.lastWordTimestamp = transcriptionEvent.last_word_end || 0;
      log(`UtteranceEnd received with last_word_end: ${this.lastWordTimestamp}`);
      
      // If we haven't already processed this utterance via speech_final
      if (!this.speechFinal && this.finalResult.trim().length > 0) {
        const trimmedResult = this.finalResult.trim();
        
        // Check if this is a duplicate of the last processed text
        if (trimmedResult === this.lastProcessedText) {
          log('Duplicate text detected, skipping processing');
          return;
        }
        
        log(`Processing utterance end: ${trimmedResult}`);
        this.isProcessingTranscription = true;
        this.clearUtteranceTimeout();
        this.latencyTracker.endTracking('Deepgram');
        this.lastProcessedText = trimmedResult;
        await this.handleTranscriptionWithInterruption(trimmedResult);
        this.finalResult = '';
        this.isProcessingTranscription = false;
      } else {
        log('Speech was already final when UtteranceEnd received');
      }
      this.speechFinal = false;
      return;
    }

    // Process regular transcription events
    if (transcriptionEvent.is_final && text.trim().length > 0) {
      this.finalResult += ` ${text}`;
      const trimmedResult = this.finalResult.trim();
      
      // If speech_final flag is set, process immediately
      if (transcriptionEvent.speech_final) {
        // Check if this is a duplicate of the last processed text
        if (trimmedResult === this.lastProcessedText) {
          log('Duplicate text detected, skipping processing');
          this.finalResult = '';
          return;
        }
        
        log(`Speech final detected: ${trimmedResult}`);
        this.isProcessingTranscription = true;
        this.clearUtteranceTimeout();
        this.speechFinal = true;
        this.latencyTracker.endTracking('Deepgram');
        this.lastProcessedText = trimmedResult;
        await this.handleTranscriptionWithInterruption(trimmedResult);
        this.finalResult = '';
        this.isProcessingTranscription = false;
      } 
      // If interruption detected, also process immediately
      else if (this.enableInterruptions && this.interruptionDetected) {
        // Check if this is a duplicate of the last processed text
        if (trimmedResult === this.lastProcessedText) {
          log('Duplicate text detected, skipping processing');
          this.finalResult = '';
          this.interruptionDetected = false;
          return;
        }
        
        log(`Interruption detected: ${trimmedResult}`);
        this.isProcessingTranscription = true;
        this.clearUtteranceTimeout();
        this.speechFinal = true;
        this.latencyTracker.endTracking('Deepgram');
        this.lastProcessedText = trimmedResult;
        await this.handleTranscriptionWithInterruption(trimmedResult);
        this.finalResult = '';
        this.isProcessingTranscription = false;
      }
      // Otherwise, log interim result but don't process yet
      else {
        this.speechFinal = false;
        log('Interim final transcription:', { text: trimmedResult });
        // Start utterance timeout for this chunk
        this.startUtteranceTimeout();
      }
    } else if (!transcriptionEvent.is_final && text.trim().length > 0) {
      log('Interim transcription:', { text: text.trim() });
      if (this.enableInterruptions && !this.isProcessingTranscription) {
        await this.checkForInterruption(text);
      }
    }

    this.transcriptionStartTime = Date.now();
  }

  private async handleTranscriptionWithInterruption(transcription: string) {
    try {
      const wasInterrupted = this.enableInterruptions && this.interruptionDetected;
      await this.handleTranscription(transcription, wasInterrupted);
      this.interruptionDetected = false;
    } catch (error) {
      log('Error in handleTranscriptionWithInterruption', { error: (error as Error).message });
      throw error;
    } finally {
      // Always reset the processing flag, even if an error occurs
      this.isProcessingTranscription = false;
      log('Reset processing flag after handling transcription');
    }
  }

  private startUtteranceTimeout() {
    if (this.utteranceTimeout) {
      clearTimeout(this.utteranceTimeout);
    }
    
    this.utteranceTimeout = setTimeout(() => {
      if (this.finalResult.trim().length > 0 && !this.speechFinal && !this.isProcessingTranscription) {
        const trimmedResult = this.finalResult.trim();
        
        // Check if this is a duplicate of the last processed text
        if (trimmedResult === this.lastProcessedText) {
          log('Duplicate text detected in timeout, skipping processing');
          this.finalResult = '';
          return;
        }
        
        log(`Utterance timeout reached, processing: ${trimmedResult}`);
        this.isProcessingTranscription = true;
        this.latencyTracker.endTracking('Deepgram');
        this.lastProcessedText = trimmedResult;
        this.handleTranscriptionWithInterruption(trimmedResult)
          .then(() => {
            this.isProcessingTranscription = false;
          })
          .catch(error => {
            log('Error in utterance timeout handler', { error: error.message });
            this.isProcessingTranscription = false;
          });
        this.finalResult = '';
        this.speechFinal = false;
      }
    }, appConfig.utterance_timeout_ms);
  }

  private clearUtteranceTimeout() {
    if (this.utteranceTimeout) {
      clearTimeout(this.utteranceTimeout);
      this.utteranceTimeout = null;
    }
  }

  private async checkForInterruption(text: string) {
    // If we're already processing a transcription or interruption, don't check again
    if (this.isProcessingTranscription || this.interruptionDetected) {
      return;
    }
    
    log('Checking for interruption', {
      enableInterruptions: this.enableInterruptions,
      textLength: text.trim().length,
      isSpeaking: isSpeaking,
      wordToInterruptEnable: appConfig.wordToInterruptEnable,
      numberofwordstointeruptenable: appConfig.numberofwordstointeruptenable,
      numberofwordstointerupt: appConfig.words_to_interrupt
    });
    
    if (this.enableInterruptions && text.trim().length > 0 && isSpeaking) {
      let shouldInterrupt = false;

      if (appConfig.wordToInterruptEnable) {
        const lowerText = text.toLowerCase();
        shouldInterrupt = appConfig.wordsToInterrupt.some(word => lowerText.includes(word.toLowerCase()));
      }

      if (appConfig.numberofwordstointeruptenable) {
        const wordCount = text.trim().split(/\s+/).length;
        shouldInterrupt = shouldInterrupt || wordCount >= appConfig.words_to_interrupt;
      }
      
      if (shouldInterrupt) {
        // Check if this is a duplicate of the last processed text
        if (text.trim() === this.lastProcessedText) {
          log('Duplicate interruption text detected, skipping');
          return;
        }
        
        log('Interruption detected', { text });
        this.isProcessingTranscription = true;
        this.interruptionDetected = true;
        this.clearUtteranceTimeout();
        
        // Send clear message immediately
        sendClearMessage(this.ws, this.streamSid);
        
        // Add delay here
        log(`Waiting for ${appConfig.clear_message_delay}ms after sending clear message`);
        await delay(appConfig.clear_message_delay);
        
        isSpeaking = false;
        this.lastProcessedText = text.trim();
        interruptionEmitter.emit('interrupt', text);
      }
    }
  }

  private startKeepAlive() {
    this.stopKeepAlive();
    this.keepAliveInterval = setInterval(() => {
      if (this.dgConnection.getReadyState() === 1) {
        this.dgConnection.send(JSON.stringify({ type: 'KeepAlive' }));
        log('Sent KeepAlive message');
      }
    }, 3000);
  }

  private stopKeepAlive() {
    if (this.keepAliveInterval) {
      clearInterval(this.keepAliveInterval);
      this.keepAliveInterval = null;
    }
  }

  private attemptReconnection() {
    if (this.isActive && this.reconnectAttempts < this.maxReconnectAttempts) {
      this.reconnectAttempts++;
      log(`Attempting to reconnect (${this.reconnectAttempts}/${this.maxReconnectAttempts})`);
      this.reconnectTimeout = setTimeout(() => {
        this.dgConnection = this.initializeConnection();
      }, 5000 * this.reconnectAttempts);
    } else {
      log('Max reconnection attempts reached or service is no longer active');
    }
  }

  public send(payload: string) {
    if (this.dgConnection.getReadyState() === 1) {
      this.latencyTracker.startTracking('Deepgram');
      this.dgConnection.send(Buffer.from(payload, 'base64'));
    } else {
      log('Cannot send: Deepgram connection is not open');
      this.latencyTracker.endTracking('Deepgram'); // End tracking if send fails
    }
  }

  public close() {
    this.isActive = false;
    this.stopKeepAlive();
    this.clearUtteranceTimeout();
    this.isProcessingTranscription = false; // Reset processing flag
    if (this.dgConnection.getReadyState() === 1) {
      this.dgConnection.finish();
      log('Deepgram connection finished');
    }
    this.latencyTracker.endTracking('Deepgram'); // Ensure tracking ends if connection closes
    log('TranscriptionService closed');
  }

  public setEnableInterruptions(enable: boolean) {
    this.enableInterruptions = enable;
    // Reset state when changing interruption settings
    this.interruptionDetected = false;
    this.isProcessingTranscription = false;
    this.lastProcessedText = '';
  }
  
  public resetProcessingFlags() {
    this.isProcessingTranscription = false;
    this.interruptionDetected = false;
    log('Transcription service processing flags reset');
  }
}

let lastResetTime = Date.now();

const resetSilenceTimer = () => {
  if (silenceTimer) {
    clearTimeout(silenceTimer);
  }
  if (!isWaitingForAudioComplete) {
    silenceTimer = setTimeout(() => {
      log('Silence timeout reached');
      if (ws && ws.readyState === WebSocket.OPEN) {
        ws.close();
      }
    }, appConfig.silenceTimeoutSeconds * 1000);
    log(`Silence timer started with timeout: ${appConfig.silenceTimeoutSeconds} seconds`);
  } else {
    log('Silence timer not started, waiting for audio_complete');
  }
};

// =====================================
// 7. Main Application Logic
// =====================================
async function startApp() {
  try {
    await fetchAssistantSettings(); // Remove the argument if it's present
  } catch (error) {
    log('Failed to fetch assistant settings:', { error: (error as Error).message });
    // You might want to implement a retry mechanism or exit the application here
    process.exit(1); // Exit the application if we can't fetch the settings
  }

  app.use(express.json());
  app.use(express.urlencoded({ extended: true }));

  app.use('/webhook', webhookRouter);

  app.post('/call/incoming', (req, res) => {
    ID = req.body.CallSid;  
    calledNumber = req.body.Called;
    From = req.body.Called;
    To = req.body.Caller;
    log('Incoming call received', { body: req.body, calledNumber, From, To });

    const twimlResponse = new twilio.twiml.VoiceResponse();
    twimlResponse.connect().stream({
      url: `wss://${process.env.SERVER_DOMAIN}:8443/call/connection`,
    });

    log('TwiML response created');

    const responseXml = twimlResponse.toString();
    res.type('text/xml').send(responseXml);
    log('TwiML response sent', { response: responseXml, calledNumber });
  });

  app.post('/call/whisper', (req, res) => {
    log('Whisper endpoint called', { body: req.body, conversationSummary });

    const twiml = new VoiceResponse();
  
    if (summaryAudioBuffer) {
      const audioHash = crypto.createHash('md5').update(summaryAudioBuffer).digest('hex');
      const audioUrl = `/audio/summary/${audioHash}.mp3`;
      twiml.play(audioUrl);
    } else {
      twiml.say("I'm sorry, but there's no summary available at the moment.");
    }
  
    //twiml.say('Do you accept the call? If yes, press number 9 on your phone keypad.');
    twiml.gather({
      numDigits: 1,
      action: '/call/whisper-handle',
      method: 'POST'
    });

    res.type('text/xml').send(twiml.toString());
    log('Whisper TwiML sent');
  });

  app.get('/audio/summary/:hash.mp3', async (req, res) => {
    const { hash } = req.params;
    
    if (summaryAudioBuffer) {
      const audioHash = crypto.createHash('md5').update(summaryAudioBuffer).digest('hex');
      
      if (hash === audioHash) {
        res.type('audio/mpeg').send(summaryAudioBuffer);
      } else {
        res.status(404).send('Audio not found');
      }
    } else {
      res.status(404).send('No audio summary available');
    }
  });

  app.post('/call/whisper-handle', (req, res) => {
    const digits = req.body.Digits;
    log('Whisper response received', { digits });

    const twiml = new VoiceResponse();
    if (digits === '9') {
      twiml.say('Connecting the call.');
      // The call will be connected automatically by Twilio
    } else {
      twiml.say('Call rejected. Goodbye.');
      twiml.hangup();
    }

    res.type('text/xml').send(twiml.toString());
  });

  app.ws('/call/connection', async (websocket: WebSocket) => {
    ws = websocket; // Assign the websocket to the global ws variable
    log('WebSocket connection established');
    let streamSid: string | undefined;
    let conversation: ConversationMessage[] = [];
    let latencyTracker = new LatencyTracker();
    let transcriptionService: TranscriptionService | null = null;
    isFirstMessage = true; // Reset isFirstMessage for each new connection
  
    const handleTranscription = async (transcription: string, wasInterrupted: boolean) => {
      log('Transcription received', { transcription, wasInterrupted });
      if (!isProcessing && streamSid) {
        isProcessing = true;
        if (isFirstMessage) {
          isFirstMessage = false;
          log('First message received');
        }
        if (silenceTimer) {
          clearTimeout(silenceTimer);
          silenceTimer = null;
          log('Silence timer stopped due to transcription');
        }
        isWaitingForAudioComplete = true;
        await handleUserInput(ws, streamSid, transcription, conversation, latencyTracker, wasInterrupted);
        isProcessing = false;
      }
    };

    // Don't start the timer when the connection is established
  
    interruptionEmitter.on('interrupt', async (interruptionText) => {
      isSpeaking = false; // Stop the current speech
      isAssistantSpeaking = false; // Update assistant speaking state
      log('is Speaking', { isSpeaking: isSpeaking, isAssistantSpeaking: isAssistantSpeaking });
      log('Interruption detected', { interruptionText });
      if (streamSid && !isProcessing) {
        isProcessing = true;
        try {
          await handleUserInput(ws, streamSid, interruptionText, conversation, latencyTracker, true);
        } catch (error) {
          log('Error handling interruption', { error: (error as Error).message });
        } finally {
          isProcessing = false;
          // Reset the transcription service's processing flags
          if (transcriptionService) {
            transcriptionService.resetProcessingFlags();
          }
        }
      } else {
        log('Skipping interruption processing - already processing or no streamSid', { 
          isProcessing, 
          hasStreamSid: Boolean(streamSid) 
        });
      }
    });

    ws.on('message', async (data) => {
      let message: TwilioMessage;

      try {
        message = JSON.parse(data.toString()) as TwilioMessage;
      } catch (error) {
        log('Error parsing WebSocket message', { error: (error as Error).message });
        return;
      }

      switch (message.event) {
        case 'connected':
          log('Connected event received');
          streamSid = message.streamSid;
          break;

        case 'start':
          if (message.start) {
            log('Start event received', { streamSid: message.start.streamSid, callSid: message.start.callSid });
            streamSid = message.start.streamSid;
            callSid = message.start.callSid;
            callStartTime = Date.now();
          
            log('Current phone number:', calledNumber || 'Not available');
          
            if (calledNumber) {
              currentAssistantSettings = await getAssistantSettings(calledNumber);
              if (currentAssistantSettings) {
                // Initialize API clients with current settings
                initializeAPIClients(currentAssistantSettings);
                
                if (currentAssistantSettings.silence_timeout) {
                  appConfig.silenceTimeoutSeconds = parseInt(currentAssistantSettings.silence_timeout);
                  log(`Applied silence timeout: ${appConfig.silenceTimeoutSeconds} seconds`);
                }
                if (currentAssistantSettings.model_provider) {
                  setLLMApiType(currentAssistantSettings.model_provider);
                  log(`Applied model provider: ${currentAssistantSettings.model_provider}`);
                }
                if (currentAssistantSettings.maximum_call_duration_enable !== undefined) {
                  appConfig.maximum_call_duration_enable = currentAssistantSettings.maximum_call_duration_enable;
                }
                if (currentAssistantSettings.maximum_call_duration) {
                  appConfig.maximum_call_duration = parseInt(currentAssistantSettings.maximum_call_duration);
                }
                log(`Applied maximum call duration settings: enabled=${appConfig.maximum_call_duration_enable}, duration=${appConfig.maximum_call_duration} seconds`);
              } else {
                log(`No specific settings found for ${calledNumber}. Using default settings.`);
              }
            } else {
              log('No phone number available. Using default settings.');
            }

            log('Full start message:', JSON.stringify(message.start, null, 2));
          
            transcriptionService = new TranscriptionService(
              handleTranscription, 
              latencyTracker, 
              appConfig.enable_interruptions,
              ws,
              streamSid
            );
            await processInitialGreeting(ws, streamSid, latencyTracker, conversation);

            // Start the maximum call duration timer
            handleMaxCallDuration(ws, streamSid);
          }
          break;

        case 'media':
          if (message.media && streamSid && transcriptionService) {
            transcriptionService.send(message.media.payload);
            // Don't reset the timer here
          }
          break;

        case 'stop':
          log('Stop event received - Call ending');
          if (transcriptionService) {
            transcriptionService.close();
            transcriptionService = null;
            log('Deepgram connection closed due to call ending');
          }
          
          // Generate call log before clearing history
          try {
            const callDuration = calculateCallDuration(); // You'll need to track call start time
            const summary = await generateConversationSummary();
            const analysis = await analyzeCallContent(chatHistory, summary, new LatencyTracker());
            
            const newCallLog: CallLog = {
              id: ID,
              phone: To,
              assistant_name: currentAssistantSettings?.assistant_name || 'Assistant',
              duration: callDuration,
              timestamp: new Date().toISOString(),
              summary,
              status: "completed", // Add logic to determine if dropped
              tags: analysis.tags || [],
              model: currentAssistantSettings?.model || appConfig.model,
              model_provider: currentAssistantSettings?.model_provider || appConfig.model_provider,
              voice_provider: currentAssistantSettings?.voice_provider || appConfig.voice_provider,
              transcriber_provider: currentAssistantSettings?.transcriber_provider || appConfig.transcriber_provider,
              transcript: conversation.map(msg => ({
                 role: msg.role,
                 message: msg.content,
                 time: msg.timestamp || new Date().toLocaleTimeString()
                })),
              analysis: {
                sentiment: analysis.sentiment || 'neutral',
                speaking_time: {
                  ai: analysis.speaking_time.ai || 0,
                  user: analysis.speaking_time.user || 0
                },
                key_topics: analysis.key_topics || [],
                tags: analysis.tags || [],
                latencies: latencyTracker.getAverages() || {}
              }
            };
            
            log('Generated Call Log:', newCallLog);
            await saveCallLog(newCallLog);
          } catch (error) {
            log('Error processing call log', { error: (error as Error).message });
          }
          
          clearChatHistory();
          latencyTracker.logSummary();
          calledNumber = ''; // Reset the calledNumber
          ws.close();
          break;

        case 'dtmf':
          if (message.dtmf) {
            log('DTMF event received', { digit: message.dtmf.digit });
          }
          break;

        case 'mark':
          if (message.mark && message.mark.name === 'audio_complete') {
            log('Received audio_complete mark from Twilio');
            isSpeaking = false;
            isAssistantSpeaking = false;
            isWaitingForAudioComplete = false;
            log('Speech playback completed', { isSpeaking: isSpeaking, isAssistantSpeaking: isAssistantSpeaking, timestamp: Date.now() });
            log('Starting silence timer');
            resetSilenceTimer();
          }
          break;

        case 'call_forwarded':
          log('Call forwarded successfully', { streamSid: message.streamSid });
          // You might want to do something here, like update the UI or log the event
          break;

        case 'call_forward_error':
          log('Error forwarding call', { streamSid: message.streamSid, error: message.error });
          // Handle the error, perhaps by informing the user or trying an alternative action
          break;

        default:
          log('Unhandled WebSocket event', { event: message.event });
      }
    });

    ws.on('error', (error) => {
      log('WebSocket error', { error: (error as Error).message });
      if (transcriptionService) {
        transcriptionService.close();
        transcriptionService = null;
        log('Deepgram connection closed due to WebSocket error');
      }
      clearChatHistory();
    });

    ws.on('close', (code, reason) => {
      log('WebSocket closed', { code, reason: reason || 'No reason provided' });
      if (transcriptionService) {
        transcriptionService.close();
        transcriptionService = null;
        log('Deepgram connection closed due to WebSocket closure');
      }
      if (silenceTimer) {
        clearTimeout(silenceTimer);
        silenceTimer = null;
      }
      clearChatHistory();
      interruptionEmitter.removeAllListeners('interrupt'); // Clean up the listener when the connection closes
    });
  });

  app.use((err: any, req: express.Request, res: express.Response, next: express.NextFunction) => {
    log('Error occurred', { error: err.stack });
    res.status(500).send('Something broke!');
  });

  app.listen(PORT, '0.0.0.0', () => {
    log('Server started', { port: PORT });
  });
}

// =====================================
// 8. Helper Functions
// =====================================

async function analyzeCallContent(transcript: ChatMessage[], summary: string, latencyTracker: LatencyTracker): Promise<CallAnalysis> {
  const sentimentPrompt = `Analyze the overall sentiment of this conversation summary and return EXACTLY one word from these options: positive, negative, or neutral. Do not return any other words, phrases or explanations.\nSummary: ${summary}`;
  
  const topicsPrompt = `Extract 3-5 key topics from this conversation. Return ONLY a comma-separated list of single words  short phrases.\nTranscript: ${transcript
   .filter(m => m.role !== 'system')  // Filter out system messages
   .map(m => m.content)
   .join('\n')}`;
  
  const tagsPrompt = `Generate 2-3 relevant tags for this conversation. Consider the type of interaction, outcome, and key points. Return ONLY a comma-separated list of single words.\nSummary: ${summary}`;

  try {
    const [sentimentResponse, topicsResponse, tagsResponse] = await Promise.all([
      getLLMResponse(sentimentPrompt, null, new LatencyTracker(), CURRENT_LLM_API),
      getLLMResponse(topicsPrompt, null, new LatencyTracker(), CURRENT_LLM_API),
      getLLMResponse(tagsPrompt, null, new LatencyTracker(), CURRENT_LLM_API)
    ]);

    log('Call Analysis Responses', {
      sentiment: sentimentResponse,
      topics: topicsResponse,
      tags: tagsResponse
    });

    const speakingTime = calculateSpeakingTime(transcript);
    log('Speaking Time Analysis', { speakingTime });

    return {
      sentiment: validateSentiment((sentimentResponse as string).toLowerCase().trim()),
      speaking_time: calculateSpeakingTime(transcript),
      key_topics: (topicsResponse as string).split(',').map(t => t.trim()),
      tags: (tagsResponse as string).split(',').map(t => t.trim()),
      latencies: latencyTracker.getAverages() // Add latencies from the tracker
    };
  } catch (error) {
    log('Error in call analysis', { error: (error as Error).message });
    return {
      sentiment: 'neutral',
      speaking_time: calculateSpeakingTime(transcript),
      key_topics: ['error_analyzing'],
      tags: ['analysis_failed'],
      latencies: {} // Add empty latencies object for error case
    };
  }
}

function validateSentiment(sentiment: string): string {
  const validSentiments = ['positive', 'negative', 'neutral'];
  return validSentiments.includes(sentiment) ? sentiment : 'neutral';
}

function calculateSpeakingTime(messages: ChatMessage[]): { ai: number; user: number } {
  // Initialize counters
  let assistantWords = 0;
  let userWords = 0;

  // Process each message
  messages.forEach(msg => {
    if (!msg.content || msg.role === 'system') return; // Skip if no content or system message
    
    // Count words and characters for user and assistant messages
    const content = msg.content.trim();
    const wordCount = content.split(/\s+/).filter(word => word.length > 0).length;
    const charCount = content.replace(/\s+/g, '').length;
    
    // Use a weighted combination of word count and character count
    const weightedCount = wordCount * 1.5 + charCount * 0.5;
    
    if (msg.role === 'assistant') {
      assistantWords += weightedCount;
    } else if (msg.role === 'user') {
      userWords += weightedCount;
    }
  });

  const totalWords = assistantWords + userWords;

  // Handle case with no words
  if (totalWords === 0) {
    return { ai: 50, user: 50 };
  }

  // Calculate percentages with proper rounding to ensure they sum to 100
  const rawAi = (assistantWords / totalWords) * 100;
  const rawUser = (userWords / totalWords) * 100;
  
  // Round to 2 decimal places
  const user = Math.round(rawAi * 100) / 100;
  const ai = Math.round(rawUser * 100) / 100;

  // Ensure percentages sum to 100
  if (ai + user !== 100) {
    const diff = 100 - (ai + user);
    // Add the difference to the larger percentage
    if (ai > user) {
      return { ai: ai + diff, user };
    } else {
      return { ai, user: user + diff };
    }
  }

  return { ai, user };
}

async function saveCallLog(newLog: CallLog): Promise<void> {
  try {
    const apiUrl = `${process.env.PERFEX_CRM_URL}/voice360/api/save_call_log`;
    const apiKey = process.env.VOICE360_API_KEY;
    
    if (!apiKey) {
      throw new Error('VOICE360_API_KEY environment variable is not set');
    }

    // Format the log data to ensure all nested objects are properly structured
    const formattedLog = {
      id: newLog.id,
      phone: newLog.phone,
      assistant_name: newLog.assistant_name,
      duration: newLog.duration,
      timestamp: newLog.timestamp,
      summary: newLog.summary,
      status: newLog.status,
      tags: Array.isArray(newLog.tags) ? newLog.tags : [],
      model: newLog.model,
      model_provider: newLog.model_provider,
      voice_provider: newLog.voice_provider,
      transcriber_provider: newLog.transcriber_provider,
      transcript: newLog.transcript.map(msg => ({
        role: msg.role,
        message: msg.message,
        time: msg.time
      })),
      analysis: {
        sentiment: newLog.analysis.sentiment,
        speaking_time: {
          assistant: newLog.analysis.speaking_time.ai,
          user: newLog.analysis.speaking_time.user
        },
        key_topics: Array.isArray(newLog.analysis.key_topics) ? newLog.analysis.key_topics : [],
        tags: Array.isArray(newLog.analysis.tags) ? newLog.analysis.tags : [],
        latencies: newLog.analysis.latencies || {}
      }
    };

    log('Attempting to save formatted call log:', formattedLog);
    
    const headers = {
      'Content-Type': 'application/json',
      'Authorization': `Bearer ${apiKey.trim()}`,
      'X-Requested-With': 'XMLHttpRequest'
    };

    const response = await fetch(apiUrl, {
      method: 'POST',
      headers: headers,
      body: JSON.stringify(formattedLog)
    });

    const responseText = await response.text();
    log('API Response:', {
      status: response.status,
      statusText: response.statusText,
      body: responseText
    });

    if (!response.ok) {
      throw new Error(`Failed to save call log: ${response.status} ${response.statusText} - ${responseText}`);
    }

    log('Call log saved successfully to Perfex CRM');
  } catch (error) {
    log('Error saving call log to Perfex CRM', { 
      error: (error as Error).message,
      stack: (error as Error).stack
    });
    
    // Fallback to local storage
    const filePath = path.join(__dirname, '..', 'data', 'call_logs.json');
    try {
      let logs: { logs: CallLog[] };
      try {
        const fileContent = await fs.readFile(filePath, 'utf8');
        logs = JSON.parse(fileContent);
      } catch (error) {
        logs = { logs: [] };
      }
      
      logs.logs.push(newLog); // Save the original log
      await fs.writeFile(filePath, JSON.stringify(logs, null, 2));
      log('Call log saved locally as fallback');
    } catch (fallbackError) {
      log('Error saving call log locally', { error: (fallbackError as Error).message });
    }
  }
}
function handleMaxCallDuration(ws: WebSocket, streamSid: string) {
  if (appConfig.maximum_call_duration_enable && appConfig.maximum_call_duration > 0) {
    setTimeout(() => {
      log('Maximum call duration reached. Ending call.');
      textToSpeechAndStream(ws, streamSid, appConfig.end_call_message, new LatencyTracker())
        .then(() => {
          if (ws.readyState === WebSocket.OPEN) {
            ws.close();
          }
        })
        .catch((error) => {
          log('Error ending call due to maximum duration', { error: (error as Error).message });
          if (ws.readyState === WebSocket.OPEN) {
            ws.close();
          }
        });
    }, appConfig.maximum_call_duration * 1000);
  }
}

function getValidOpenAIVoice(voice: string): OpenAIVoice {
  const validVoices: OpenAIVoice[] = ["alloy", "echo", "fable", "onyx", "nova", "shimmer"];
  if (validVoices.includes(voice as OpenAIVoice)) {
    return voice as OpenAIVoice;
  }
  console.warn(`Invalid OpenAI voice: ${voice}. Defaulting to 'alloy'.`);
  return "alloy";
}

async function processInitialGreeting(ws: WebSocket, streamSid: string, latencyTracker: LatencyTracker, conversation: ConversationMessage[]) {
  try {
    isProcessing = true;
    log('Processing initial greeting');
    
    // Initialize chat history with the combined system prompt
    initializeChatHistory();
    
    // Create timestamp for the initial greeting
    const timestamp = new Date().toLocaleTimeString();
    
    // Add initial "Hi" to conversation array
    conversation.push({ 
      role: 'user',
      content: "Hi",
      timestamp: timestamp
    });
     
    latencyTracker.startTracking('LLM');
    const openAIResponse = await getLLMResponse("Hi", null, latencyTracker, CURRENT_LLM_API);
    latencyTracker.endTracking('LLM');
    log('LLM response received', { response: openAIResponse });

    // Add assistant's response to conversation array
    const responseContent = typeof openAIResponse === 'string' 
      ? openAIResponse 
      : await new Promise<string>((resolve, reject) => {
          let content = '';
          openAIResponse.on('data', chunk => content += chunk);
          openAIResponse.on('end', () => resolve(content));
          openAIResponse.on('error', reject);
        });

    conversation.push({
      role: 'assistant',
      content: responseContent,
      timestamp: new Date().toLocaleTimeString()
    });

    await textToSpeechAndStream(ws, streamSid, openAIResponse, latencyTracker);
  } catch (error) {
    log('Error in processInitialGreeting', { error: (error as Error).message });
    await textToSpeechAndStream(ws, streamSid, "I'm sorry, but I encountered an error. Please try again later.", latencyTracker);
  } finally {
    isProcessing = false;
  }
}

async function handleUserInput(ws: WebSocket, streamSid: string, userInput: string, conversation: ConversationMessage[], latencyTracker: LatencyTracker, wasInterrupted: boolean) {
  latencyTracker.startTracking('Total');
  try {
    log('Handling user input', { userInput });
    
    // Stop the silence timer when user input is received
    isWaitingForAudioComplete = true;
    resetSilenceTimer();
    
    let contextualPrompt = userInput;
    if (appConfig.enable_interruptions && isAssistantSpeaking) {
      contextualPrompt = `The user interrupted with: "${userInput}". Please acknowledge the interruption politely and address their new input.`;
    }
    
    conversation.push({ 
      role: 'user', 
      content: contextualPrompt,
      timestamp: new Date().toLocaleTimeString()
    });
    log('Current conversation context', { conversation });

    // Implement LLM request delay
    log(`Waiting for ${appConfig.llm_request_delay} seconds before sending request to LLM`);
    await delay(appConfig.llm_request_delay * 1000);

    latencyTracker.startTracking('LLM');
    const openAIResponse = await getLLMResponse(contextualPrompt, currentThreadId, latencyTracker, CURRENT_LLM_API);
    latencyTracker.endTracking('LLM');

    log('LLM response received', { response: openAIResponse });
    if (typeof openAIResponse === 'string') {
      conversation.push({ role: 'assistant', content: openAIResponse });

      // Check for call forwarding phrase
      log('Checking for call forwarding phrase', { 
        call_forwarding_enable: appConfig.call_forwarding_enable, 
        response: openAIResponse.toLowerCase() 
      });
      if (appConfig.call_forwarding_enable && openAIResponse.toLowerCase().includes("forwarding your call")) {
        log('Call forwarding phrase detected');
        // Generate the conversation summary
        const summary = await generateConversationSummary();
        conversationSummary = summary; // Store in the global variable
        // Generate the audio only when forwarding
        log('Generating audio for summary');
        summaryAudioBuffer = await generateAudioBuffer(summary);
        if (callSid) {
          await forwardCall(ws, streamSid, callSid);
        } else {
          log('Error: CallSid is not available for forwarding');
        }
        return; // Exit the function after forwarding the call
      }

      // Check for SMS booking text phrase
      if (appConfig.enable_sms_booking_text && openAIResponse.toLowerCase().includes("send sms booking text")) {
        log('SMS booking text phrase detected');
        await sendSMSBookingText(ws, streamSid);
        return; // Exit the function after sending SMS, without sending the LLM response to TTS
      }
    }

    // Implement response delay
    log(`Waiting for ${appConfig.response_delay} seconds before responding`);
    await delay(appConfig.response_delay * 1000);

    isAssistantSpeaking = true;
    latencyTracker.startTracking('TTS');
    await textToSpeechAndStream(ws, streamSid, openAIResponse, latencyTracker);
    latencyTracker.endTracking('TTS');

    latencyTracker.endTracking('Total');

    log('Latency Breakdown', {
      deepgram: latencyTracker.getLatency('Deepgram'),
      llm: latencyTracker.getLatency('LLM'),
      tts: latencyTracker.getLatency('TTS'),
      total: latencyTracker.getLatency('Total')
    });

    // Don't reset the silence timer here, it will be reset when 'audio_complete' is received
  } catch (error) {
    log('Error in handleUserInput', { error: (error as Error).message });
    await textToSpeechAndStream(ws, streamSid, "I'm sorry, but I encountered an error. Could you please try again?", latencyTracker);
    isAssistantSpeaking = false;
    resetSilenceTimer();
  }
}

async function getLLMResponse(
  input: string, 
  threadId: string | null, 
  latencyTracker: LatencyTracker, 
  apiType: LLMApiType = CURRENT_LLM_API
): Promise<string | Readable> {
  if (Array.isArray(input)) {
    input = input.join(' ');
  }
  log('Sending message to LLM', { input, threadId, apiType });

  latencyTracker.startTracking('LLM');
  try {
    // Use the stored settings instead of fetching again
    const assistantSettings = currentAssistantSettings;
    if (apiType === LLMApiType.OpenAIAssistant) {
      let thread;
      if (threadId) {
        thread = await openai.beta.threads.retrieve(threadId);
        await openai.beta.threads.messages.create(threadId, {
          role: 'user',
          content: input
        });
      } else {
        thread = await openai.beta.threads.create({
          messages: [{ role: 'user', content: input }]
        });
        currentThreadId = thread.id;
      }

      log('Using thread', { threadId: thread.id });

      const stream = new Readable({
        read() {}
      });

      const runStream = await openai.beta.threads.runs.createAndStream(thread.id, {
        assistant_id: process.env.OPENAI_ASSISTANT_ID!,
      });

      (async () => {
        for await (const chunk of runStream) {
          const event = chunk as AssistantStreamEvent;
          if (event.event === 'thread.message.delta') {
            const delta = event.data.delta;
            if (delta.content && delta.content[0].type === 'text') {
              const chunkValue = delta.content[0].text.value;
              log('Received chunk from OpenAI Assistant', { chunk: chunkValue });
              stream.push(chunkValue);
            }
          }
        }
        stream.push(null);
      })();

      return stream;
    } else if (apiType === LLMApiType.OpenAIChat) {
      chatHistory.push({ role: 'user', content: input });

      switch (COMMUNICATION_METHOD) {
        case CommunicationMethod.WebSocket: {
          const stream = new Readable({
            read() {}
          });

          const chatStream = await openai.chat.completions.create({
            model: currentAssistantSettings.model || "gpt-4o-mini-2024-07-18",
            messages: chatHistory,
            stream: true,
            temperature: currentAssistantSettings.temperature || 1,
            max_tokens: currentAssistantSettings.max_tokens || 256,
            top_p: currentAssistantSettings.top_p || 1,
            frequency_penalty: 0,
            presence_penalty: 0,
          });

          let assistantResponse = '';

          (async () => {
            for await (const chunk of chatStream) {
              if (chunk.choices[0]?.delta?.content) {
                const chunkValue = chunk.choices[0].delta.content;
                log('Received chunk from OpenAI Chat', { chunk: chunkValue });
                stream.push(chunkValue);
                assistantResponse += chunkValue;
              }
            }
            chatHistory.push({ role: 'assistant', content: assistantResponse });
            
            if (chatHistory.length > 10) {
              chatHistory = [
                chatHistory[0],
                ...chatHistory.slice(-9)
              ];
            }
            
            stream.push(null);
          })();

          return stream;
        }
        case CommunicationMethod.Streaming:
        case CommunicationMethod.NonStreaming:
        default: {
          latencyTracker.startTracking('LLM');    
          const completion = await openai.chat.completions.create({
            model: currentAssistantSettings.model || "gpt-4o-mini-2024-07-18",
            messages: chatHistory,
            temperature: currentAssistantSettings.temperature || 1,
            max_tokens: currentAssistantSettings.max_tokens || 256,
            top_p: currentAssistantSettings.top_p || 1,
            frequency_penalty: 0,
            presence_penalty: 0,
          });
          latencyTracker.endTracking('LLM');
          
          const assistantResponse = completion.choices[0]?.message?.content || '';
          chatHistory.push({ 
            role: 'assistant', 
            content: assistantResponse,
            timestamp: new Date().toLocaleTimeString()
          });
          
          if (chatHistory.length > 10) {
            chatHistory = [
              chatHistory[0],
              ...chatHistory.slice(-9)
            ];
          }

          return assistantResponse;
        }
      }
    } else if (apiType === LLMApiType.Groq) {
      chatHistory.push({ role: 'user', content: input });

      switch (COMMUNICATION_METHOD) {
        case CommunicationMethod.WebSocket: {
          const stream = new Readable({
            read() {}
          });

          const chatStream = await groq.chat.completions.create({
            model: currentAssistantSettings.model || "llama3-groq-70b-8192-tool-use-preview",
            messages: chatHistory,
            stream: true,
            temperature: currentAssistantSettings.temperature || 0.65,
            max_tokens: currentAssistantSettings.max_tokens || 256,
            top_p: currentAssistantSettings.top_p || 1,
          });

          let assistantResponse = '';

          (async () => {
            for await (const chunk of chatStream) {
              if (chunk.choices[0]?.delta?.content) {
                const chunkValue = chunk.choices[0].delta.content;
                log('Received chunk from Groq', { chunk: chunkValue });
                stream.push(chunkValue);
                assistantResponse += chunkValue;
              }
            }
            chatHistory.push({ role: 'assistant', content: assistantResponse });
            
            if (chatHistory.length > 10) {
              chatHistory = [
                chatHistory[0],
                ...chatHistory.slice(-9)
              ];
            }
            
            stream.push(null);
          })();

          return stream;
        }
        case CommunicationMethod.Streaming:
        case CommunicationMethod.NonStreaming:
        default: {
          latencyTracker.startTracking('LLM');    
          const completion = await groq.chat.completions.create({
            model: currentAssistantSettings.model || "llama3-groq-70b-8192-tool-use-preview",
            messages: chatHistory,
            stream: false,
            temperature: currentAssistantSettings.temperature || 0.65,
            max_tokens: currentAssistantSettings.max_tokens || 256,
            top_p: currentAssistantSettings.top_p || 1,
          });
          latencyTracker.endTracking('LLM');
          
          const assistantResponse = completion.choices[0]?.message?.content || '';
          chatHistory.push({ role: 'assistant', content: assistantResponse });
          
          if (chatHistory.length > 10) {
            chatHistory = [
              chatHistory[0],
              ...chatHistory.slice(-9)
            ];
          }

          return assistantResponse;
        }
      }
    } else {
      throw new Error(`Unsupported LLM API type: ${apiType}`);
    }
  } catch (error) {
    latencyTracker.endTracking('LLM');
    throw error;
  }
}

// =====================================
// 9. Text-to-Speech and Streaming Functions
// =====================================
async function textToSpeechAndStream(ws: WebSocket, streamSid: string, text: string | Readable, latencyTracker: LatencyTracker) {
  let interruptionOccurred = false;
  let audioCompleteSent = false;
  const handleInterruption = (ws: WebSocket, streamSid: string) => {
    if (isSpeaking) {
      interruptionOccurred = true;
      log('Interruption detected, sending clear message');
      sendClearMessage(ws, streamSid);
      // Note: We'll set isSpeaking to false after receiving confirmation from Twilio
    }
  };
  try {
    log('Initializing TTS API');
    
    let textToConvert: string;
    if (text instanceof Readable) {
      textToConvert = await new Promise<string>((resolve, reject) => {
        let content = '';
        text.on('data', (chunk) => content += chunk);
        text.on('end', () => resolve(content));
        text.on('error', reject);
      });
    } else {
      textToConvert = text;
    }
    log('Converting text to speech', { textLength: textToConvert.length });

    // Set up interruption listener just before starting to speak
    interruptionEmitter.on('interrupt', () => handleInterruption(ws, streamSid));
    isSpeaking = true;
    log('Is Speaking', { isSpeaking: isSpeaking });
    log('Starting speech', { timestamp: Date.now() });
    
    latencyTracker.startTracking('TTS');
    
    switch (appConfig.voice_provider.toLowerCase()) {
	 case 'elevenlabs':
			switch (COMMUNICATION_METHOD) {
			  case CommunicationMethod.WebSocket: {
				const elevenlabsWs = new WebSocket(`wss://api.elevenlabs.io/v1/text-to-speech/${appConfig.voice_id}/stream-input?model_id=eleven_flash_v2&output_format=ulaw_8000&optimize_streaming_latency=3`);

				elevenlabsWs.on('open', async () => {
				  log('ElevenLabs WebSocket opened');
				  elevenlabsWs.send(JSON.stringify({
					"text": " ",
					"voice_settings": {
					  "stability": appConfig.stability,
					  "similarity_boost": appConfig.clarity_boost
					},
					"xi_api_key": currentAssistantSettings.provider_settings.elevenlabs_api_key
				  }));

				  const sendChunk = (chunk: string) => {
					log('Sending chunk to ElevenLabs', { chunk });    
					elevenlabsWs.send(JSON.stringify({ 
					  "text": chunk,
					}));
				  };

				  const processText = async () => {
					if (text instanceof Readable) {
					  for await (const chunk of text_chunker(createAsyncIterableFromReadable(text))) {
						if (interruptionOccurred) break;
						sendChunk(chunk);
					  }
					} else {
					  for await (const chunk of text_chunker(createAsyncIterableFromString(text))) {
						if (interruptionOccurred) break;
						sendChunk(chunk);
					  }
					}
					elevenlabsWs.send(JSON.stringify({ "flush": true }));
					elevenlabsWs.send(JSON.stringify({ "text": "" }));
					log('Sending Final Messages to ElevenLabs'); 
				  };

				  processText();
				});

				let isFirstChunk = true;

				elevenlabsWs.on('message', (data) => {
				  const message = JSON.parse(data.toString());
				  if (message.audio) {
					const audioChunk = Buffer.from(message.audio, 'base64');
					log('Received audio chunk from ElevenLabs', { 
					  chunkSizeBytes: audioChunk.length,
					  firstFewBytes: audioChunk.slice(0, 10).toString('hex')
					});
					if (!interruptionOccurred) {
					  sendMediaMessage(ws, streamSid, audioChunk);
					}
					
					if (isFirstChunk) {
					  isFirstChunk = false;
					  log('Received first audio chunk from ElevenLabs');
					}
				  } else if (message.isFinal) {
					log('Received final message from ElevenLabs');
					if (!interruptionOccurred && !audioCompleteSent) {
					  audioCompleteSent = true;
					  sendMarkMessage(ws, streamSid, 'audio_complete');
					}
				  }
				});

				elevenlabsWs.on('error', (error) => {
				  log('ElevenLabs WebSocket error', { error: error.message });
				  if (!audioCompleteSent) {
					audioCompleteSent = true;
					sendMarkMessage(ws, streamSid, 'audio_complete');
				  }
				});

				elevenlabsWs.on('close', () => {
				  log('ElevenLabs WebSocket closed');
				});

				await new Promise((resolve) => {
				  elevenlabsWs.on('close', resolve);
				});

				break;
			  }

			  case CommunicationMethod.Streaming:
				const audioStream = await elevenlabs.generate({
				  text: textToConvert,
				  voice: appConfig.voice_id,
				  model_id: modelId,
				  voice_settings: {
					stability: appConfig.stability,
					similarity_boost: appConfig.clarity_boost
				  },
				  output_format: 'ulaw_8000',
				  optimize_streaming_latency: '3'
				});
			  
				log('Starting audio streaming to Twilio');
				
				let isFirstChunk = true;
				audioStream.on('data', (chunk) => {
				  if (isFirstChunk) {
					latencyTracker.endTracking('TTS');
					isFirstChunk = false;
				  }
				  log('Received audio chunk from ElevenLabs', { size: chunk.length });
				  if (!interruptionOccurred) {
					sendMediaMessage(ws, streamSid, chunk);
				  }
				});
			  
				audioStream.on('end', () => {
				  log('ElevenLabs audio stream ended');
				  if (!interruptionOccurred && !audioCompleteSent) {
					audioCompleteSent = true;
					sendMarkMessage(ws, streamSid, 'audio_complete');
				  }
				});
			  
				await new Promise((resolve) => audioStream.on('end', resolve));
				break;

			  case CommunicationMethod.NonStreaming:
			  default:
				const audioResponse = await elevenlabs.textToSpeech.convert(appConfig.voice_id, {
				  text: textToConvert,
				  model_id: modelId,
				  voice_settings: {
					stability: 0.5,
					similarity_boost: 0.75
				  },
				  output_format: 'ulaw_8000',
				  optimize_streaming_latency: '3'
				});
				
				latencyTracker.endTracking('TTS');

				log('Received response from ElevenLabs', { 
				  responseType: typeof audioResponse, 
				  isBuffer: Buffer.isBuffer(audioResponse),
				  isArrayBuffer: audioResponse instanceof ArrayBuffer,
				  isReadable: audioResponse instanceof Readable
				});

				let audioBuffer: Buffer;
				if (audioResponse instanceof ArrayBuffer) {
				  audioBuffer = Buffer.from(audioResponse);
				} else if (Buffer.isBuffer(audioResponse)) {
				  audioBuffer = audioResponse;
				} else if (audioResponse instanceof Readable) {
				  audioBuffer = await new Promise<Buffer>((resolve, reject) => {
					const chunks: Buffer[] = [];
					audioResponse.on('data', (chunk) => chunks.push(Buffer.from(chunk)));
					audioResponse.on('end', () => resolve(Buffer.concat(chunks)));
					audioResponse.on('error', reject);
				  });
				} else {
				  throw new Error(`Unexpected response type from ElevenLabs: ${typeof audioResponse}`);
				}

				log('Processed audio from ElevenLabs', { size: audioBuffer.length });
				if (!interruptionOccurred) {
				  sendMediaMessage(ws, streamSid, audioBuffer);
				  audioCompleteSent = true;
				  sendMarkMessage(ws, streamSid, 'audio_complete');
				}
				break;
			}
			break;

		  case 'openai':
			log('Starting OpenAI TTS process');
			const mp3Audio = await openai.audio.speech.create({
			  model: "tts-1",
			  voice: getValidOpenAIVoice(appConfig.voice_id),
			  input: textToConvert,
			  response_format: "mp3",
			});

			log('Converting OpenAI response to audio stream');
			
			// Convert the response to a readable stream
			const audioStream = new Readable({
			  read() {}
			});

			// Process the audio data
			(async () => {
			  const audioResponse = mp3Audio as unknown as { arrayBuffer: () => Promise<ArrayBuffer> };
			  const buffer = Buffer.from(await audioResponse.arrayBuffer());
			  audioStream.push(buffer);
			  audioStream.push(null); // Signal the end of the stream
			})();

			log('Starting FFmpeg conversion process');

			const ffmpegProcess = spawn(ffmpegPath, [
			  '-i', 'pipe:0',
			  '-acodec', 'pcm_mulaw',
			  '-ar', '8000',
			  '-f', 'mulaw',
			  'pipe:1'
			]);

			audioStream.pipe(ffmpegProcess.stdin);

			let isFirstChunk = true;
			ffmpegProcess.stdout.on('data', (chunk) => {
			  if (isFirstChunk) {
				latencyTracker.endTracking('TTS');
				isFirstChunk = false;
			  }
			  log('Sending converted audio chunk to Twilio', { size: chunk.length });
			  if (!interruptionOccurred) {
				sendMediaMessage(ws, streamSid, chunk);
			  }
			});

			ffmpegProcess.stderr.on('data', (data) => {
			  log('FFmpeg stderr:', data.toString());
			});

			ffmpegProcess.on('close', (code) => {
			  log('FFmpeg process closed', { code });
			  if (!interruptionOccurred && !audioCompleteSent) {
				audioCompleteSent = true;
				sendMarkMessage(ws, streamSid, 'audio_complete');
			  }
			});

			await new Promise((resolve) => ffmpegProcess.on('close', resolve));
			break;
		  
		case 'deepgram':
	  log('Starting Deepgram TTS process');
	  
	  try {
		const response = await deepgram.speak.request(
		  { text: textToConvert },
		  {
			model: appConfig.voice_id,
			encoding: "mulaw",
			sample_rate: 8000
		  }
		);
		const stream = await response.getStream();
		
		if (stream) {
		  log('Received audio stream from Deepgram');
		  latencyTracker.endTracking('TTS');
		  const reader = stream.getReader();
		  
		  while (true) {
			const { done, value } = await reader.read();
			
			if (done) {
			  log('Deepgram stream completed naturally');
			  break;
			}
			sendMediaMessage(ws, streamSid, Buffer.from(value));
		  }
		  
		  log('Sending audio_complete mark');
		  sendMarkMessage(ws, streamSid, 'audio_complete');
		} else {
		  throw new Error('No audio stream received from Deepgram');
		}
	  } catch (deepgramError) {
		log('Error in Deepgram TTS', { error: (deepgramError as Error).message });
		throw deepgramError;
	  }
	  break;
      case 'rime_ai':
        log('Starting RimeAI TTS process');
        try {
          const response = await fetch('https://users.rime.ai/v1/rime-tts', {
            method: 'POST',
            headers: {
              Accept: 'audio/x-mulaw',
              Authorization: `Bearer ${currentAssistantSettings.provider_settings.rime_ai_api_key}`,
              'Content-Type': 'application/json'
            },
            body: JSON.stringify({
              speaker: appConfig.voice_id,
              text: textToConvert,
              modelId: RIME_AI_CONFIG.modelId,
              speedAlpha: appConfig.speed,
              reduceLatency: RIME_AI_CONFIG.reduceLatency
            })
          });

          if (!response.ok) {
            throw new Error(`RimeAI API error: ${response.status} ${response.statusText}`);
          }

          if (!response.body) {
            throw new Error('No response body received from RimeAI');
          }

          const reader = response.body.getReader();
          log('Received audio stream from RimeAI');
          latencyTracker.endTracking('TTS');

          while (true) {
            const { done, value } = await reader.read();
            
            if (done) {
              log('RimeAI stream completed naturally');
              break;
            }

            sendMediaMessage(ws, streamSid, Buffer.from(value));
            
            // Small delay to allow for interruption checking
            await new Promise(resolve => setTimeout(resolve, 50));

            if (interruptionOccurred) {
              log('Interruption occurred, stopping RimeAI audio stream');
              await reader.cancel();
              break;
            }
          }
          
          if (!interruptionOccurred) {
            log('Sending audio_complete mark');
            sendMarkMessage(ws, streamSid, 'audio_complete');
          }
        } catch (rimeAIError) {
          log('Error in RimeAI TTS', { error: (rimeAIError as Error).message });
          throw rimeAIError;
        }
        break;
    }
  } catch (error) {
    log('Error in textToSpeechAndStream', { error: (error as Error).message, stack: (error as Error).stack });
    const errorMessage = "I'm sorry, but I encountered an error while processing the audio. Please try again.";
    const errorBuffer = Buffer.from(errorMessage, 'utf-8');
    sendMediaMessage(ws, streamSid, errorBuffer);
    sendMarkMessage(ws, streamSid, 'audio_complete');
    latencyTracker.endTracking('TTS');
  } finally {
    log('Speech completed', { timestamp: Date.now() });
  }
}

async function* text_chunker(chunks: AsyncIterable<string>): AsyncGenerator<string> {
  const splitters = [".", ",", "?", "!", ";", ":", "—", "-", "(", ")", "[", "]", "}", " "];
  let buffer = "";

  for await (const text of chunks) {
    if (splitters.some(splitter => buffer.endsWith(splitter))) {
      yield buffer + " ";
      buffer = text;
    } else if (splitters.some(splitter => text.startsWith(splitter))) {
      yield buffer + text[0] + " ";
      buffer = text.slice(1);
    } else {
      buffer += text;
    }
  }

  if (buffer) {
    yield buffer + " ";
  }
}

function createAsyncIterableFromReadable(readable: Readable): AsyncIterable<string> {
  return {
    [Symbol.asyncIterator]() {
      return {
        next() {
          return new Promise((resolve) => {
            readable.once('data', (chunk) => {
              resolve({ value: chunk.toString(), done: false });
            });
            readable.once('end', () => {
              resolve({ value: undefined, done: true });
            });
          });
        },
      };
    },
  };
}

function createAsyncIterableFromString(str: string): AsyncIterable<string> {
  return {
    [Symbol.asyncIterator]() {
      let done = false;
      return {
        next() {
          if (done) {
            return Promise.resolve({ done: true, value: undefined });
          }
          done = true;
          return Promise.resolve({ done: false, value: str });
        },
      };
    },
  };
}

function sendMediaMessage(ws: WebSocket, streamSid: string, payload: Buffer) {
  const message = {
    event: 'media',
    streamSid: streamSid,
    media: {
      payload: payload.toString('base64')
    },
    sequenceNumber: (mediaSequenceNumber++).toString()
  };
  const messageString = JSON.stringify(message);
  log('Sending media message to Twilio', { 
    streamSid,
    sequenceNumber: message.sequenceNumber,
    payloadSize: payload.length, 
    messageSize: messageString.length
  });
  ws.send(messageString);
}

function sendMarkMessage(ws: WebSocket, streamSid: string, name: string) {
  const message = {
    event: 'mark',
    streamSid: streamSid,
    mark: {
      name: name
    }
  };
  log('Sending mark message', { message });
  ws.send(JSON.stringify(message));
}

// =====================================
// 10. Webhook Handler
// =====================================
app.post('/webhook', express.json(), async (req, res) => {
  const authHeader = req.headers.authorization;
  if (!authHeader || !authHeader.startsWith('Bearer ')) {
    log('Webhook request received without proper authorization');
    return res.status(401).json({ error: 'Unauthorized' });
  }

  const token = authHeader.split(' ')[1];
  if (token !== process.env.VOICE360_API_KEY) {
    log('Webhook request received with invalid API key');
    return res.status(401).json({ error: 'Invalid API key' });
  }

  try {
    await fetchAssistantSettings();
    log('Cache updated successfully via webhook');
    res.json({ message: 'Cache updated successfully' });
  } catch (error) {
    log('Error updating cache via webhook', { error: (error as Error).message });
    res.status(500).json({ error: 'Failed to update cache' });
  }
});

// =====================================
// 11. Error Handling and Process Event Listeners
// =====================================
process.on('unhandledRejection', (reason, promise) => {
  log('Unhandled Rejection', { reason });
});

process.on('uncaughtException', (error) => {
  log('Uncaught Exception', { error: (error as Error).message, stack: (error as Error).stack });
  process.exit(1);
});

// Application Start
log('Starting application');
startApp();

// Send SMS Booking
async function sendSMSBookingText(ws: WebSocket, streamSid: string) {
  try {
    if (!From || !To) {
      throw new Error('From or To number is not available');
    }

    // Use Twilio client to send SMS
    await twilioClient.messages.create({
      body: appConfig.sms_booking_text,
      from: From,
      to: To
    });

    log('SMS booking text sent successfully');

    // Inform the user that the SMS has been sent
    await textToSpeechAndStream(ws, streamSid, "Excellent! I’ve just sent a text with the booking link to your phone number, and I’ll stay on the line in case you need any help along the way!.", new LatencyTracker());
  } catch (error) {
    log('Error sending SMS booking text', { error: (error as Error).message });
    await textToSpeechAndStream(ws, streamSid, "I'm sorry, but there was an error sending the SMS. Please try again later.", new LatencyTracker());
  }
}
